#include <stdio.h>
#include <stdlib.h>
#include <i86.h>
#include <dos.h>
#include <string.h>

void disable(void);
#pragma aux disable = \
  "cli" \
  modify exact [];

void enable(void);
#pragma aux enable = \
  "sti" \
  modify exact [];

void call_dos_idle(void);
#pragma aux call_dos_idle = \
  "int 0x28";

int dos_close(unsigned handle);
#pragma aux dos_close = \
  "mov ah, 0x3e" \
  "int 0x21" \
  parm [bx] value [ax];

unsigned get_dos_dmabuffer(void *ret);
#pragma aux get_dos_dmabuffer = \
  "mov ah, 0x6a" \
  "mov cx, 6" \
  "int 0x21" \
  parm [di] value [ax];

unsigned char __based(__segname("_CODE")) error_code = 0;

unsigned get_kernel_cs(void);
#pragma aux get_kernel_cs = \
  "mov ah, 0x18" \
  "int 0x21" \
  value [dx] modify [ax];

#include "errors.c"

unsigned char call_critical_error(unsigned drive, unsigned error_code);
#pragma aux call_critical_error = \
  "or ah, 0x7E" \
  "int 0x2C" \
  parm [ax] [di] value [al];

inline unsigned char critical_error(unsigned char drive, unsigned char error, unsigned char write)
{
  error = bios2doserror(error);
  return call_critical_error(drive | (write << 8), error);
}

unsigned char bios_error_code(void);
#pragma aux bios_error_code = \
  "mov ah, 1" \
  "int 0x13" \
  value [al];

void driver_break(void);
#pragma aux driver_break = \
  "int 0x2B" \
  modify [ax];


union ipv4_address
{
  unsigned long address;
  unsigned char n[4];
};

struct sockaddr_in
{
  unsigned sin_family;
  unsigned sin_port;
  union ipv4_address ip_address;
  char unused[8];
};

struct netdisk
{
  unsigned long offset;
  unsigned long last_access;
  int socket;
  char filename[13];
};

struct netdisk __far * __based(__segname("_CODE")) disks;
unsigned __based(__segname("_CODE")) disk_count;

void debug(unsigned char chr);
#pragma aux debug = \
  "mov ah, 0x0E" \
  "mov bx, 0x0007" \
  "int 0x10" \
  parm [al] modify [bx];

unsigned current_psp(void);
#pragma aux current_psp = \
  "mov ah, 0x51" \
  "int 0x21" \
  value [bx] modify [ax];

void __far *dos_alloc(unsigned bytes);
#pragma aux dos_alloc = \
  "push cx" \
  "mov cl, 4" \
  "add bx, 0xF" \
  "shr bx, cl" \
  "pop cx" \
  "mov ah, 0x48" \
  "int 0x21" \
  "jnc end" \
  "xor ax, ax" \
  "end:" \
  "xor dx, dx" \
  "xchg ax, dx" \
  parm [bx] value [dx ax];

void __far dos_free(unsigned seg);
#pragma aux dos_free = \
  "mov ah, 0x49" \
  "mov es, bx" \
  "int 0x21" \
  parm [bx] modify [es ax];

void tcp_service(void);
#pragma aux tcp_service = \
  "xor ax, ax" \
  "int 0x61" \
  modify exact [ax];

int socket(unsigned family, unsigned type);
#pragma aux socket = \
  "mov ax, 9" \
  "int 0x61" \
  parm [bx] [cx] value [ax];

int connect(int socket, void __far *);
#pragma aux connect = \
  "mov ax, 11" \
  "int 0x61" \
  value [ax] parm [bx] [es dx];

int send(int socket, void __far *buff, unsigned count);
#pragma aux send = \
  "mov ax, 3" \
  "int 0x61" \
  value [ax] parm [bx] [es dx] [cx];

int recv(int socket, void __far *buff, unsigned count);
#pragma aux recv = \
  "mov ax, 2" \
  "int 0x61" \
  value [ax] parm [bx] [es dx] [cx];

int close_socket(int socket);
#pragma aux close_socket =  \
  "mov ax, 4" \
  "int 0x61" \
  parm [bx] value [ax];

unsigned tcp_error(void);
#pragma aux tcp_error = \
  "mov ax, 6" \
  "int 0x61" \
  value [ax];

void copymem(void __far *dst, void __far *src, unsigned count);
#pragma aux copymem = \
  "push ds" \
  "cld" \
  "mov ds, dx" \
  "rep movsb" \
  "pop ds" \
  parm [es di] [dx si] [cx];

int send_udp(int s, char __far *buff, unsigned count)
{
  int retval;

  retval = send(s, buff, count);
  tcp_service();

  return retval;
}

#define CACHE_SIZE 16
#define BLOCK_SIZE 512
#define CACHE_TIMEOUT 5*18

unsigned long bios_clock_ticks(void);
#pragma aux bios_clock_ticks = \
  "xor ah, ah" \
  "int 0x1A" \
  value [cx dx] modify [ax];

struct block_cache
{
  unsigned long block;
  unsigned seg;
  unsigned refs;
  unsigned char drive;
};
struct block_cache __based(__segname("_CODE")) block_caches[CACHE_SIZE];

unsigned kernel_allocate_usercache(unsigned, void*);
#pragma aux kernel_allocate_usercache = \
  "push ds" \
  "mov ax, cs" \
  "mov ds, ax" \
  "mov ax, 0x6d02" \
  "int 0x21" \
  "pop ds" \
  "jnc noerror" \
  "mov ax, -1" \
  "noerror:" \
  parm [bx] [di] modify [ax] value [ax];

static int allocate_cache(unsigned char drive)
{
  unsigned n;
  int rc = 0;
  unsigned lr = -1;

  for(n=CACHE_SIZE;n--;)
  {
    if(block_caches[n].drive == 0xFF)
    {
      rc = n;
      break;
    }
    if(block_caches[n].refs < lr)
    {
      lr = block_caches[n].refs;
      rc = n;
    }
  }
  disable();
  if(block_caches[rc].drive != 0xFF)
    goto allocated;

  n = kernel_allocate_usercache(BLOCK_SIZE>>4, &block_caches[rc].drive);
  if(n != -1)
  {
    block_caches[rc].seg = n;
    allocated:
    enable();
    block_caches[rc].drive = drive;
    block_caches[rc].refs = 16;
    return rc;
  }
  else
  {
    enable();
    return -1;
  }
}

static void invalidate_cache(unsigned char drive)
{
  unsigned n;
  disks[drive].last_access = bios_clock_ticks();
  for(n=CACHE_SIZE;n--;)
    if(block_caches[n].drive == drive)
    {
      block_caches[n].drive = -1;
      dos_free(block_caches[n].seg);
    }
}

static int find_cache_block(unsigned char drive, unsigned long block)
{
  int n;
  int rc = -1;
  unsigned long clk = bios_clock_ticks();
  unsigned long last_access = disks[drive].last_access;

  disks[drive].last_access = clk;

  if(clk - last_access > CACHE_TIMEOUT)
  {
    invalidate_cache(drive);
    return -1;
  }

  for(n=CACHE_SIZE;n--;)
    if(block_caches[n].drive == drive && block_caches[n].block == block)
    {
      block_caches[n].refs+=2;
      rc = n;
    }
    else
      if(block_caches[n].refs)
        block_caches[n].refs--;
  return rc;
}

inline void init_cache(void)
{
  unsigned n;
  for(n=CACHE_SIZE;n--;)
    block_caches[n].drive = -1;
}

/*static void save_to_cache(unsigned char drive, unsigned long block, void __far *data)
{
  int c = find_cache_block(drive, block);

  if(c == -1)
  {
    c = allocate_cache(drive);
    if(c == -1) return;
  }
  else return;
  block_caches[c].block = block;
  copymem(MK_FP(block_caches[c].seg, 0), data, BLOCK_SIZE);
}*/


volatile unsigned __based(__segname("_CODE")) clock_ticks;
void __far timer_rtn();
extern void __far * __based(__segname("_CODE")) old_timer_rtn;

#define TIMEOUT 18

int recv_udp(int s, char __far *buff, unsigned count)
{
  int retval;

  clock_ticks = 0;

  do
  {
    tcp_service();
    retval = recv(s, buff, count);
    if(!retval)
      call_dos_idle();
    else break;
  }
  while(clock_ticks < TIMEOUT);

  return retval;
}

#define RETRY_COUNT 3

#define REQ_HDR_LEN 1+12+8+2
#define PAYLOAD_LEN 256

#define ANS_HDR_LEN 1+12+8+2
#define ANSWER_LEN ANS_HDR_LEN+PAYLOAD_LEN

static unsigned __far netdisk_read(unsigned, unsigned char __far*, unsigned);
#pragma aux netdisk_read value [ax] parm [bx] [dx ax] [cx] modify exact [];
static unsigned __far netdisk_read(unsigned devnum, unsigned char __far *buff, unsigned count)
{
  unsigned n;
  int rc;
  int sd = disks[devnum].socket;
  char dbuff[6];
  unsigned reqcnt;
  unsigned recvcnt = 0;
  unsigned retval = 0;
  unsigned retries;
  unsigned long block;
  int c;
  char __far *ptr;
  unsigned short_offset = disks[devnum].offset & (BLOCK_SIZE-1);
  unsigned cache_read_count;
  unsigned cache_read_bytes = 0;
  unsigned copy_count;
  unsigned long old_offset;
  #define req (*(unsigned char __far**)&dbuff[2])
  #define buff_size (*(unsigned *)&dbuff[0])

  if(rc = get_dos_dmabuffer(dbuff))
  {
    error_code = rc;
    return retval;
  }

  while(count)
  {
    block = (disks[devnum].offset + retval) >> 9;
    retries = RETRY_COUNT;

    disable();
    if((c = find_cache_block(devnum, block)) != -1)
    {
      ptr = MK_FP(block_caches[c].seg, 0);
      copy_count = count;
      if(copy_count > BLOCK_SIZE-short_offset)
        copy_count = BLOCK_SIZE-short_offset;
      copymem(&buff[retval], &ptr[short_offset], count);
      enable();
      disks[devnum].offset += copy_count;
      count -= copy_count;
      retval += copy_count;
      short_offset = 0;
      cache_read_bytes += copy_count;
      cache_read_count -= copy_count;
      continue;
    }

    c = allocate_cache(devnum);

    old_offset = disks[devnum].offset;

    if(c != -1)
    {
      ptr = MK_FP(block_caches[c].seg, 0);
      block_caches[c].block = block;
      // align offset to block
      disks[devnum].offset &= ~(unsigned long)(BLOCK_SIZE-1);
      cache_read_count = BLOCK_SIZE;
    }
    else
    {
      ptr = 0;
      short_offset = 0;
      cache_read_count = count;
    }

    do
    {
      if(ptr)
        reqcnt = PAYLOAD_LEN;
      else
      {
        reqcnt = count;
        if(reqcnt > PAYLOAD_LEN)
          reqcnt = PAYLOAD_LEN;
      }

      req[0] = 'R';
      copymem(&req[1], &disks[devnum].filename, 12);
      *(unsigned long __far*)&req[1+12] = disks[devnum].offset+cache_read_bytes;
      *(unsigned long __far*)&req[1+12+4] = 0;
      *(unsigned __far*)&req[1+12+8] = reqcnt;

      rc = send_udp(sd, req, REQ_HDR_LEN);
      if(rc < REQ_HDR_LEN)
        goto error;

      rc = recv_udp(sd, req, reqcnt+ANS_HDR_LEN);
      if(rc <= 0)
        if(retries--)
          continue;
        else
          goto error;

      if(req[0] == 'E')
      {
        error:
//        error_code = 0x0E;
//        return retval;
        switch(critical_error(devnum, error_code = 0x15, 0))
        {
          case 0: // ignore
            error_code = 0;
            return retval;
          case 1: // retry
            error_code = 0;
            retries = RETRY_COUNT;
            continue;
          case 2: // abort
            driver_break();
          case 3: // fail
            return retval;
        }
      }

      recvcnt = *(unsigned __far*)&req[1+12+8];

      if(count)
        if(short_offset < PAYLOAD_LEN)
        {
          copy_count = recvcnt - short_offset;
          if(copy_count > count)
            copy_count = count;
          copymem(&buff[retval], &req[ANS_HDR_LEN+short_offset], copy_count);
          count -= copy_count;
          retval += copy_count;
          short_offset = 0;
        }
        else
          short_offset -= PAYLOAD_LEN;
 
      if(ptr)
      {
        disable();
        if(block_caches[c].drive != 0xFF)
          copymem(&ptr[cache_read_bytes], &req[ANS_HDR_LEN], recvcnt);
        enable();
      }
      cache_read_bytes += recvcnt;
      cache_read_count -= recvcnt;
    }
    while(cache_read_count);

    disks[devnum].offset = old_offset + retval;
  }

  return retval;
}


static unsigned __far netdisk_write(unsigned devnum, unsigned char __far *buff, unsigned count);
#pragma aux netdisk_write parm [bx] [dx ax] [cx] value [ax] modify exact [];
static unsigned __far netdisk_write(unsigned devnum, unsigned char __far *buff, unsigned count)
{
  unsigned n;
  int rc;
  int sd = disks[devnum].socket;
  char dbuff[6];
  unsigned reqcnt;
  unsigned sendcnt = 0;
  unsigned retval = 0;
  unsigned retries = RETRY_COUNT;
  int c;
  unsigned long block;
  unsigned short_offset;
  unsigned cache_write_count;
  #define req (*(unsigned char __far**)&dbuff[2])
  #define buff_size (*(unsigned *)&dbuff[0])

  rc = get_dos_dmabuffer(dbuff);
  if(rc)
  {
    error_code = rc;
    return retval;
  }

  while(count)
  {
    block = disks[devnum].offset >> 9;

    reqcnt = count;
    if(reqcnt > PAYLOAD_LEN)
      reqcnt = PAYLOAD_LEN;

    req[0] = 'W';
    copymem(&req[1], disks[devnum].filename, 12);
    *(unsigned long __far*)&req[1+12] = disks[devnum].offset;
    *(unsigned long __far*)&req[1+12+4] = 0;
    *(unsigned __far*)&req[1+12+8] = reqcnt;
    copymem(&req[REQ_HDR_LEN], &buff[retval], reqcnt);

    rc = send_udp(sd, req, reqcnt+REQ_HDR_LEN);
    if(rc <= 0)
      goto error;

    rc = recv_udp(sd, req, ANS_HDR_LEN);
    if(rc <= 0)
      if(retries--)
        continue;
      else
        goto error;

    if(req[0] == 'E')
    {
      error:
//      error_code = 0x0E;
//      return retval;
      switch(critical_error(devnum, error_code = 0x0E, 1))
      {
        case 0:
          error_code = 0;
          return retval;
        case 1:
          error_code = 0;
          retries = RETRY_COUNT;
          continue;
        case 2:
          driver_break();
        case 3:
          return retval;
      }
    }

    sendcnt = *(unsigned __far*)&req[1+12+8];

    if((c = find_cache_block(devnum, block)) != -1)
    {
      short_offset = disks[devnum].offset & (BLOCK_SIZE-1);
      cache_write_count = sendcnt;
      if(cache_write_count > BLOCK_SIZE-short_offset)
        cache_write_count = BLOCK_SIZE-short_offset;
      copymem(MK_FP(block_caches[c].seg, short_offset), &buff[retval],
              cache_write_count);
    }


    count -= sendcnt;
    retval += sendcnt;
    disks[devnum].offset += sendcnt;
  }

  return retval;
}

static unsigned long __far netdisk_seek(unsigned devnum, unsigned long count, unsigned origin);
#pragma aux netdisk_seek parm [bx] [dx ax] [cx] value [dx ax] modify exact [dx ax];
static unsigned long __far netdisk_seek(unsigned devnum, unsigned long count, unsigned origin)
{
  if(!origin)
    return disks[devnum].offset = count;
  else if(origin == 1)
    return disks[devnum].offset += count;
  else if(origin == 2)
    return -1;
}

static int __far netdisk_open(unsigned devnum, unsigned mode);
#pragma aux netdisk_open parm [bx] [cx] value [ax];
static int __far netdisk_open(unsigned devnum, unsigned mode)
{
  if(devnum >= disk_count)
  {
    error_code = 2;
    return -1;
  }
  return 0;
}

unsigned __far get_error(unsigned devnum);
#pragma aux get_error parm [bx] value [ax];
unsigned __far get_error(unsigned devnum)
{
  unsigned retval = error_code;
  error_code = 0;
  return retval;
}

int __far driver_command(unsigned, unsigned short __far*);
#pragma aux driver_command parm [bx] [dx ax] value [ax];
int __far driver_command(unsigned devnum, unsigned short __far *cmd)
{
  return 0;
}

unsigned long __far driver_ioctl(unsigned, unsigned, unsigned char __far*, unsigned);
#pragma aux driver_ioctl parm [bx] [cx] [dx ax] [si] value [dx ax];

#define CHAR_DEVICE 0x80L
#define NO_EOF_ON_INPUT 0x40L
#define BINARY_MODE 0x20L

unsigned long __far driver_ioctl(unsigned devnum, unsigned count,
                                 unsigned char __far *buff, unsigned function)
{
  unsigned return_word;

//  eof = (serial_cb[devnum].write_offset == serial_cb[devnum].read_offset);
  // this device never EOFs
  switch(function & 0xFF)
  {
    case 0: case 1:
      return_word = CHAR_DEVICE | BINARY_MODE;
      return ((long)return_word << 16) | function;
    case 6:
      return ((long)FP_SEG(buff) << 16) | (function | 0xFF);
    case 7:
      return ((long)FP_SEG(buff) << 16) | (function | 0xFF);
  }
}


int install_driver(char __far *, void __far *);
#pragma aux install_driver = \
  "mov ah, 0x6A" \
  "xor cx, cx" \
  "push ds" \
  "push es" \
  "pop ds" \
  "int 0x21" \
  "pop ds" \
  "jnc noerror" \
  "mov ax, -1" \
  "jmp end" \
  "noerror:" \
  "xor ax, ax" \
  "end:" \
  parm [dx bx] [es di] value [ax] modify [cx];

  void __far * __based(__segname("_CODE")) function_table[] =
  {
    netdisk_read,
    netdisk_write,
    netdisk_seek, // seek
    netdisk_open, // open
    0, // close
    get_error, // get_error_code
    driver_command, // send_command
    driver_ioctl // ioctl
  };


void set_memory_allocation_strategy(unsigned strategy);
#pragma aux set_memory_allocation_strategy = \
  "mov ax, 0x5801" \
  "int 0x21" \
  parm [bx];

int main()
{
  unsigned env_seg;
  unsigned char __far *psp_ptr = MK_FP(current_psp(), 0);
  int rc;
  unsigned n;
  char drivername[] = "ND";

  puts("Loading NETDISK driver v1.4.");

//  set_memory_allocation_strategy(2);
  rc = initialize_driver();
//  set_memory_allocation_strategy(0);

  if(rc)
  {
    return rc;
  }

  if(!install_driver(drivername, function_table))
  {
    puts("Done, going resident...");

    env_seg = *(unsigned __far*)&psp_ptr[0x2C];
    dos_free(env_seg);

    for(n=3;n--;)
      dos_close(n);

    old_timer_rtn = _dos_getvect(0x1C);
    _dos_setvect(0x1C, timer_rtn);

    _dos_keep(0, 0x10 + ((FP_OFF(main) + 0xF) >> 4));
  }
  puts("Error");
  return 1;
}

unsigned long resolve_dns(char *address)
{
  int s;
  unsigned long retval;
  struct sockaddr_in sockaddr;
  void *ptr1 = &sockaddr;

  printf("Resolving DNS address of %s...\n", address);
  // use tcp stack's built-in dns resolver...
  __asm
  {
    // zero error code
    mov ax, 6
    int 0x61

    loop1:
    mov ax, 1
    mov si, address
    mov cx, 1
    // just ignore BX == port and create an UDP socket...
    int 0x61
    mov s, ax

    mov ax, 6
    int 0x61
    cmp ax, 0x102 // is error code == ENOTCONN?
    jnz end_loop1 // if not, exit the loop

    xor ax, ax
    int 0x61

    jmp loop1

    end_loop1:

    // do we have an error?
    test ax, ax
    // if yes, the loop could not be resolved
    jnz end
    // else get the remote address from the socket with getpeername()
    mov ax, 12
    mov bx, s
    push ss
    pop es
    mov dx, ptr1
    mov cx, 16 // size of struct sockaddr_in
    int 0x61
    push ax // save the return value

    // and close the socket
    mov ax, 4
    mov bx, s
    int 0x61

    pop ax  // did getpeername return with non-zero?
    mov s, ax

    end:
  }

  if(s) return -1; // if getpeername returned with an error, return -1

  printf("IP address: %u.%u.%u.%u\n", sockaddr.ip_address.n[0],
         sockaddr.ip_address.n[1], sockaddr.ip_address.n[2],
         sockaddr.ip_address.n[3]);
  return sockaddr.ip_address.address;
}

int parse_config_line(char *line, unsigned id)
{
  char *ptr = strchr(line, ':');
  char *address = line;
  struct sockaddr_in addr;
  int rc;
  unsigned long ip_address;
  unsigned port;

  if(!ptr) return 1;
  *ptr = 0;
  if((ip_address = resolve_dns(line)) == -1)
  {
    printf("Could not resolve DNS for the host %s\n", line);
    return 4;
  }
  line = ptr+1;
  ptr = strchr(line, '/');
  if(!ptr) return 2;
  *ptr = 0;
  port = atoi(line);
  line = ptr+1;
  if(!strlen(line))
    return 3;
  ptr = strchr(line, '\n');
  if(ptr) *ptr = 0;
  _fstrcpy(disks[id].filename, line);

  disks[id].socket = socket(0, 1);
  if(disks[id].socket == -1)
  {
    puts("Cannot create socket! Out of memory?");
    return 5;
  }
  addr.ip_address.address = ip_address;
  addr.sin_port = (port >> 8) | (port << 8);
  addr.sin_family = 0;  

  rc = connect(disks[id].socket, &addr);  
  if(rc == -1)
  {
    puts("Socket could not be connected.");
    return 6;
  }

  printf("ND%u is connected to %s:%u/%s\n", id, address, port, line);

  return 0;
}

int parse_config(void)
{
  static char diskstr[] = "DISK=";
  char buff[100];
  int retval = 0;
  int rc;

  FILE *fp = fopen("netdisk.ini", "r");
  if(!fp) return -1;

  for(retval=0;fgets(buff, sizeof(buff), fp);)
    if(!memcmp(buff, diskstr, sizeof(diskstr)-1))
      retval++;

  printf("%u disks configured.\n", retval);

  set_memory_allocation_strategy(2);
  disks = dos_alloc(retval*sizeof(struct netdisk));
  set_memory_allocation_strategy(0);
  if(!disks)
    return -2;
  _fmemset(disks, 0, retval*sizeof(struct netdisk));

  rewind(fp);

  for(retval=0;fgets(buff, sizeof(buff), fp);)
    if(!memcmp(buff, diskstr, sizeof(diskstr)-1))
    {
      rc = parse_config_line(&buff[sizeof(diskstr)-1], retval);
      if(rc)
        return -3;
      retval++;
    }

  fclose(fp);

  return retval;
}

int initialize_driver(void)
{
  disk_count = parse_config();
  
  if(disk_count == -1)
  {
    puts("Cannot open NETDISK.INI.");
    return 1;
  }
  if(!disk_count)
  {
    puts("No netdisks configured.");
    return 1;
  }
  if(disk_count == -2)
  {
    puts("Out of memory.");
    return 1;
  }
  if(disk_count == -3)
  {
    puts("Syntax error in config file.");
    return 1;
  }

  init_cache();
}
